"""
Module for serialization/deserialization and handling of KNX addresses.

The module can handle:

* physical addresses of devices.
* (logical) group addresses.

The module supports all different writings of group addresses:

* 3rn level: "1/2/3"
* 2nd level: "1/2"
* Free format: "123"
"""
from enum import Enum
from xknx.exceptions import CouldNotParseAddress

class AddressType(Enum):
    """Enum class for different type of addresses."""
    PHYSICAL = 1
    GROUP = 2

class AddressFormat(Enum):
    """Enum class for different writing of addresses."""
    LEVEL2 = 1
    LEVEL3 = 2
    FREE = 3

class Address:
    """Class for handling KNX pyhsical and group addresses."""

    def __init__(self, address=0, address_type=None):
        """Initialize Address class."""
        self.raw = None
        self.address_format = None
        self._set(address, address_type)

    def _set(self, address, address_type):

        self.address_type = \
            address_type if address_type is not None \
            else self._detect_address_type(address)

        if address is None:
            self.raw = 0
            self.address_format = AddressFormat.LEVEL3

        elif isinstance(address, Address):
            self.raw = address.raw
            self.address_format = address.address_format

        elif isinstance(address, str):
            if self.address_type == AddressType.PHYSICAL:
                self._set_str_physical(address)
            else:
                self._set_str_group(address)

        elif isinstance(address, int):
            self._set_int(address)

        elif isinstance(address, tuple):
            self._set_tuple(address)

        else:
            raise TypeError()

    def _set_str_physical(self, address):
        parts = address.split(".")
        if any(not part.isdigit() for part in parts):
            raise CouldNotParseAddress(address)
        if len(parts) != 3:
            raise CouldNotParseAddress(address)
        main = int(parts[0])
        middle = int(parts[1])
        sub = int(parts[2])
        if main > 15:
            raise CouldNotParseAddress(address)
        if middle > 15:
            raise CouldNotParseAddress(address)
        if sub > 255:
            raise CouldNotParseAddress(address)
        self.raw = (main<<12) +  (middle<<8) + sub
        self.address_format = AddressFormat.LEVEL3


    def _set_str_group(self, address):
        parts = address.split("/")

        if any(not part.isdigit() for part in parts):
            raise CouldNotParseAddress(address)
        if len(parts) == 1:
            self._set_int(int(parts[0]))
        elif len(parts) == 2:
            self._set_str_group_level2(parts)
        elif len(parts) == 3:
            self._set_str_group_level3(parts)
        else:
            raise CouldNotParseAddress(address)

    def _set_str_group_level2(self, parts):
        main = int(parts[0])
        sub = int(parts[1])
        if main > 31:
            raise CouldNotParseAddress(parts)
        if sub > 2047:
            raise CouldNotParseAddress(parts)
        self.raw = (main<<11) + sub
        self.address_format = AddressFormat.LEVEL2

    def _set_str_group_level3(self, parts):
        main = int(parts[0])
        middle = int(parts[1])
        sub = int(parts[2])
        if main > 31:
            raise CouldNotParseAddress(parts)
        if middle > 7:
            raise CouldNotParseAddress(parts)
        if sub > 255:
            raise CouldNotParseAddress(parts)
        self.raw = (main<<11) +  (middle<<8) + sub
        self.address_format = AddressFormat.LEVEL3

    def _set_tuple(self, address):
        if len(address) != 2 \
                or any(not isinstance(byte, int) for byte in address) \
                or any(byte < 0 for byte in address) \
                or any(byte > 255 for byte in address):
            raise CouldNotParseAddress(address)

        self._set_int(address[0] * 256 + address[1])

    def _set_int(self, raw):
        if not isinstance(raw, int):
            raise CouldNotParseAddress(raw)
        if raw > 65535:
            raise CouldNotParseAddress(raw)
        self.raw = raw
        self.address_format = AddressFormat.FREE

    def str(self):
        """Return the address in the written string representation."""
        if self.address_type == AddressType.PHYSICAL:
            return self._str_physical()
        elif self.address_format == AddressFormat.FREE:
            return self._str_free()
        elif self.address_format == AddressFormat.LEVEL2:
            return self._str_level2()
        elif self.address_format == AddressFormat.LEVEL3:
            return self._str_level3()
        else:
            raise TypeError()

    def _str_free(self):
        """Return the address in the written string representation in free format."""
        return '{0}'.format((self.raw & 65535))

    def _str_level2(self):
        """Return the address in the written string representation in 2nd level format."""
        return '{0}/{1}'.format(
            ((self.raw >> 11) & 15),
            (self.raw & 4095))

    def _str_level3(self):
        """Return the address in the written string representation 3rd level format."""
        return '{0}/{1}/{2}'.format(
            ((self.raw >> 11) & 31),
            ((self.raw >> 8) & 7),
            (self.raw & 255))

    def _str_physical(self):
        """Return the address in the written string representation in the format for physical addresses."""
        return '{0}.{1}.{2}'.format(
            ((self.raw >> 12) & 15),
            ((self.raw >> 8) & 15),
            (self.raw & 255))

    @staticmethod
    def _detect_address_type(address):
        """Return the type of address from the style of writing."""
        # Physical addresses have either be specified explicitely or
        # in the correct notation. As default an address is a group address.
        if isinstance(address, str) and "." in address:
            return AddressType.PHYSICAL
        elif isinstance(address, Address):
            return address.address_type
        else:
            return AddressType.GROUP

    def to_knx(self):
        """Serialize to KNX/IP raw data."""
        return (self.raw >> 8) & 255, self.raw & 255

    def __str__(self):
        """Return object as readable string."""
        return '<Address str="{0}" />'.format(self.str())

    def __eq__(self, other):
        """Equal operator."""
        if other is None:
            return False
        if not isinstance(other, Address):
            raise TypeError()
        return self.raw == other.raw



"""
Device is the base class for all implemented devices (e.g. Lights/Switches/Sensors.)

It provides basis functionality for reading the state from the KNX bus.
"""

import asyncio
from xknx.knx import Telegram

class Device:
    """Base class for devices."""

    def __init__(self, xknx, name, device_updated_cb=None):
        """Initialize Device class."""
        self.xknx = xknx
        self.name = name
        self.device_updated_cbs = []
        if device_updated_cb is not None:
            self.register_device_updated_cb(device_updated_cb)

    def register_device_updated_cb(self, device_updated_cb):
        """Register device updated callback."""
        self.device_updated_cbs.append(device_updated_cb)

    def unregister_device_updated_cb(self, device_updated_cb):
        """Unregister device updated callback."""
        self.device_updated_cbs.remove(device_updated_cb)

    @asyncio.coroutine
    def after_update(self):
        """Method is be called by derived classes after the internal state has been changed."""
        for device_updated_cb in self.device_updated_cbs:
            # pylint: disable=not-callable
            yield from device_updated_cb(self)

    @asyncio.coroutine
    def sync(self, wait_for_result=True):
        """Read state of device from KNX bus."""
        self.xknx.logger.debug("Sync %s", self.name)
        for group_address in self.state_addresses():
            from xknx.core import ValueReader
            value_reader = ValueReader(self.xknx, group_address)
            if wait_for_result:
                telegram = yield from value_reader.read()
                if telegram is not None:
                    yield from self.process(value_reader.received_telegram)
                else:
                    self.xknx.logger.warning("Could not read value of %s %s", self, group_address)
            else:
                yield from value_reader.send_group_read()

    @asyncio.coroutine
    def send(self, group_address, payload=None):
        """Send payload as telegram to KNX bus."""
        telegram = Telegram()
        telegram.group_address = group_address
        telegram.payload = payload
        yield from self.xknx.telegrams.put(telegram)

    def state_addresses(self):
        """Return group addresses which should be requested to sync state."""
        # pylint: disable=no-self-use
        return []

    @asyncio.coroutine
    def process(self, telegram):
        """Process incoming telegram."""
        pass

    def get_name(self):
        """Return name of device."""
        return self.name

    # pylint: disable=invalid-name
    @asyncio.coroutine
    def do(self, action):
        """Abstract method for executing 'do' commands (e.g. do('position:10') ). Has to be implemented by derived class."""
        pass



"""Basis class for all KNX/IP bodies."""

class KNXIPBody():
    """Basis class for all KNX/IP bodies."""

    def calculated_length(self):
        """Get length of KNX/IP body."""
        pass

    def from_knx(self, raw):
        """Parse/deserialize from KNX/IP raw data."""
        pass

    def to_knx(self):
        """Serialize to KNX/IP raw data."""
        pass



"""Module for handling some global settings of XKNX."""
from xknx.knx import Address, AddressType

class Globals():
    """Class for handling some global settings of XKNX."""
    # pylint: disable=too-few-public-methods

    def __init__(self):
        """Initialize Globals class."""

        # own_address is the own group address of the XKNX daemon
        self.own_address = Address("15.15.250", AddressType.PHYSICAL)

        #own_ip is the own ip address of the daemon. Needed if multiple
        #network interfaces are available to detect on which the multicast
        #daemon should listen
        self.own_ip = None



"""
Module for reading configfiles (xknx.yaml).

* it will parse the given file
* and add the found devices to the devies vector of XKNX.
"""

import yaml
from xknx.knx import Address, AddressType
from xknx.devices import BinarySensor, Climate, Time, Light, Switch, Cover, Sensor

class Config:
    """Class for parsing xknx.yaml."""

    def __init__(self, xknx):
        """Initialize Config class."""
        self.xknx = xknx

    def read(self, file='xknx.yaml'):
        """Read config."""
        self.xknx.logger.debug("Reading %s", file)
        with open(file, 'r') as filehandle:
            doc = yaml.load(filehandle)
            self.parse_general(doc)
            self.parse_groups(doc)

    def parse_general(self, doc):
        """Parse the general section of xknx.yaml."""
        if "general" in doc:
            if "own_address" in doc["general"]:
                self.xknx.globals.own_address = \
                    Address(doc["general"]["own_address"],
                            AddressType.PHYSICAL)
            if "own_ip" in doc["general"]:
                self.xknx.globals.own_ip = doc["general"]["own_ip"]

    def parse_groups(self, doc):
        """Parse the group section of xknx.yaml."""
        for group in doc["groups"]:
            if group.startswith("light"):
                self.parse_group_light(doc["groups"][group])
            elif group.startswith("switch"):
                self.parse_group_switch(doc["groups"][group])
            elif group.startswith("cover"):
                self.parse_group_cover(doc["groups"][group])
            elif group.startswith("climate"):
                self.parse_group_climate(doc["groups"][group])
            elif group.startswith("time"):
                self.parse_group_time(doc["groups"][group])
            elif group.startswith("sensor"):
                self.parse_group_sensor(doc["groups"][group])
            elif group.startswith("binary_sensor"):
                self.parse_group_binary_sensor(doc["groups"][group])

    def parse_group_light(self, entries):
        """Parse a light section of xknx.yaml."""
        for entry in entries:
            light = Light.from_config(self.xknx,
                                      entry,
                                      entries[entry])
            self.xknx.devices.add(light)

    def parse_group_switch(self, entries):
        """Parse a switch section of xknx.yaml."""
        for entry in entries:
            switch = Switch.from_config(self.xknx,
                                        entry,
                                        entries[entry])
            self.xknx.devices.add(switch)

    def parse_group_binary_sensor(self, entries):
        """Parse a binary_sensor section of xknx.yaml."""
        for entry in entries:
            binary_sensor = BinarySensor.from_config(self.xknx,
                                                     entry,
                                                     entries[entry])
            self.xknx.devices.add(binary_sensor)

    def parse_group_cover(self, entries):
        """Parse a cover section of xknx.yaml."""
        for entry in entries:
            cover = Cover.from_config(self.xknx,
                                      entry,
                                      entries[entry])
            self.xknx.devices.add(cover)

    def parse_group_climate(self, entries):
        """Parse a climate section of xknx.yaml."""
        for entry in entries:
            climate = Climate.from_config(self.xknx,
                                          entry,
                                          entries[entry])
            self.xknx.devices.add(climate)

    def parse_group_time(self, entries):
        """Parse a time section of xknx.yaml."""
        for entry in entries:
            time = Time.from_config(self.xknx,
                                    entry,
                                    entries[entry])
            self.xknx.devices.add(time)

    def parse_group_sensor(self, entries):
        """Parse a sensor section of xknx.yaml."""
        for entry in entries:
            sensor = Sensor.from_config(self.xknx,
                                        entry,
                                        entries[entry])
            self.xknx.devices.add(sensor)



"""
Module for managing a light via KNX.

It provides functionality for

* switching light 'on' and 'off'.
* setting the brightness.
* reading the current state from KNX bus.
"""
import asyncio
from xknx.knx  import Address, DPTBinary, DPTArray
from xknx.exceptions import CouldNotParseTelegram
from .device import Device

class Light(Device):
    """Class for managing a light."""

    def __init__(self,
                 xknx,
                 name,
                 group_address_switch=None,
                 group_address_switch_state=None,
                 group_address_brightness=None,
                 group_address_brightness_state=None,
                 device_updated_cb=None):
        """Initialize Light class."""
        # pylint: disable=too-many-arguments
        Device.__init__(self, xknx, name, device_updated_cb)
        if isinstance(group_address_switch, (str, int)):
            group_address_switch = Address(group_address_switch)
        if isinstance(group_address_switch_state, (str, int)):
            group_address_switch_state = Address(group_address_switch_state)
        if isinstance(group_address_brightness, (str, int)):
            group_address_brightness = Address(group_address_brightness)
        if isinstance(group_address_brightness_state, (str, int)):
            group_address_brightness_state = Address(group_address_brightness_state)

        self.group_address_switch = group_address_switch
        self.group_address_switch_state = group_address_switch_state
        self.group_address_brightness = group_address_brightness
        self.group_address_brightness_state = group_address_brightness_state
        self.state = False
        self.brightness = 0
        self.supports_dimming = \
            group_address_brightness is not None


    @classmethod
    def from_config(cls, xknx, name, config):
        """Initialize object from configuration structure."""
        group_address_switch = \
            config.get('group_address_switch')
        group_address_switch_state = \
            config.get('group_address_switch_state')
        group_address_brightness = \
            config.get('group_address_brightness')
        group_address_brightness_state = \
            config.get('group_address_brightness_state')

        return cls(xknx,
                   name,
                   group_address_switch=group_address_switch,
                   group_address_switch_state=group_address_switch_state,
                   group_address_brightness=group_address_brightness,
                   group_address_brightness_state=group_address_brightness_state)


    def has_group_address(self, group_address):
        """Test if device has given group address."""
        return (self.group_address_switch == group_address) or \
               (self.group_address_switch_state == group_address) or \
               (self.group_address_brightness == group_address) or \
               (self.group_address_brightness_state == group_address)


    def __str__(self):
        """Return object as readable string."""
        if not self.supports_dimming:
            return '<Light name="{0}" ' \
                    'group_address_switch="{1}" ' \
                    'group_address_switch_state="{2}" ' \
                    'state="{3}" />' \
                    .format(
                        self.name,
                        self.group_address_switch,
                        self.group_address_switch_state,
                        self.state)

        return '<Light name="{0}" ' \
                'group_address_switch="{1}" ' \
                'group_address_switch_state="{2}" ' \
                'group_address_brightness="{3}" ' \
                'group_address_brightness_state="{4}" ' \
                'state="{5}" brightness="{6}" />' \
                .format(
                    self.name,
                    self.group_address_switch,
                    self.group_address_switch_state,
                    self.group_address_brightness,
                    self.group_address_brightness_state,
                    self.state,
                    self.brightness)

    @asyncio.coroutine
    def _set_internal_state(self, state):
        """Set the internal state of the device. If state was changed after update hooks are executed."""
        if state != self.state:
            self.state = state
            yield from self.after_update()

    @asyncio.coroutine
    def _set_internal_brightness(self, brightness):
        """Set the internal brightness of the device. If state was changed after update hooks are executed."""
        if brightness != self.brightness:
            self.brightness = brightness
            yield from self.after_update()

    @asyncio.coroutine
    def set_on(self):
        """Switch light on."""
        yield from self.send(self.group_address_switch, DPTBinary(1))
        yield from self._set_internal_state(True)

    @asyncio.coroutine
    def set_off(self):
        """Switch light off."""
        yield from self.send(self.group_address_switch, DPTBinary(0))
        yield from self._set_internal_state(False)

    @asyncio.coroutine
    def set_brightness(self, brightness):
        """Set brightness of light."""
        if not self.supports_dimming:
            return
        yield from self.send(self.group_address_brightness, DPTArray(brightness))
        yield from self._set_internal_brightness(brightness)

    @asyncio.coroutine
    def do(self, action):
        """Method for executing 'do' commands."""
        if action == "on":
            yield from self.set_on()
        elif action == "off":
            yield from self.set_off()
        elif action.startswith("brightness:"):
            yield from self.set_brightness(int(action[11:]))
        else:
            self.xknx.logger.warning("Could not understand action %s for device %s", action, self.get_name())

    def state_addresses(self):
        """Return group addresses which should be requested to sync state."""
        state_address_switch = \
            self.group_address_switch_state or \
            self.group_address_switch
        state_addresses = [state_address_switch]
        if self.supports_dimming:
            state_address_brightness = \
                self.group_address_brightness_state or \
                self.group_address_brightness
            state_addresses.append(state_address_brightness)
        return state_addresses

    @asyncio.coroutine
    def process(self, telegram):
        """Process incoming telegram."""
        if telegram.group_address == self.group_address_switch or \
            telegram.group_address == self.group_address_switch_state:
            yield from self._process_state(telegram)
        elif self.supports_dimming and \
                telegram.group_address == self.group_address_brightness or \
                telegram.group_address == self.group_address_brightness_state:
            yield from self._process_brightness(telegram)

    @asyncio.coroutine
    def _process_brightness(self, telegram):
        """Process incoming telegram for brightness state."""
        if not isinstance(telegram.payload, DPTArray) or \
            len(telegram.payload.value) != 1:
            raise CouldNotParseTelegram()

        yield from self._set_internal_brightness(telegram.payload.value[0])

    @asyncio.coroutine
    def _process_state(self, telegram):
        """Process incoming telegram for on/off state."""
        if not isinstance(telegram.payload, DPTBinary):
            raise CouldNotParseTelegram()
        if telegram.payload.value == 0:
            yield from self._set_internal_state(False)
        elif telegram.payload.value == 1:
            yield from self._set_internal_state(True)
        else:
            raise CouldNotParseTelegram()

    def __eq__(self, other):
        """Equal operator."""
        return self.__dict__ == other.__dict__



"""Abstraction to send DisconnectRequest and wait for DisconnectResponse."""
from xknx.knxip import KNXIPServiceType, KNXIPFrame, HPAI,\
    DisconnectResponse
from .request_response import RequestResponse

class Disconnect(RequestResponse):
    """Class to send a DisconnectRequest and wait for a DisconnectResponse."""

    def __init__(self, xknx, udp_client, communication_channel_id):
        """Initialize Disconnect class."""
        self.xknx = xknx
        self.udp_client = udp_client

        super(Disconnect, self).__init__(self.xknx, self.udp_client, DisconnectResponse)

        self.communication_channel_id = communication_channel_id


    def create_knxipframe(self):
        """Create KNX/IP Frame object to be sent to device."""
        (local_addr, local_port) = self.udpclient.getsockname()
        knxipframe = KNXIPFrame()
        knxipframe.init(KNXIPServiceType.DISCONNECT_REQUEST)
        knxipframe.body.communication_channel_id = \
            self.communication_channel_id
        knxipframe.body.control_endpoint = HPAI(
            ip_addr=local_addr, port=local_port)

        return knxipframe



"""
Module for managing the climate within a room.

* It reads/listens to a temperature address from KNX bus.
* Manages and sends the desired setpoint to KNX bus.
"""
import time
import datetime
import asyncio
from xknx.knx import Address, DPTArray, DPTTemperature
from xknx.exceptions import CouldNotParseTelegram
from .device import Device


class Climate(Device):
    """Class for managing the climate."""

    def __init__(self,
                 xknx,
                 name,
                 group_address_temperature=None,
                 group_address_setpoint=None,
                 device_updated_cb=None):
        """Initialize Climate class."""
        # pylint: disable=too-many-arguments
        Device.__init__(self, xknx, name, device_updated_cb)
        if isinstance(group_address_temperature, (str, int)):
            group_address_temperature = Address(group_address_temperature)
        if isinstance(group_address_setpoint, (str, int)):
            group_address_setpoint = Address(group_address_setpoint)

        self.group_address_temperature = group_address_temperature
        self.group_address_setpoint = group_address_setpoint
        self.last_set = None
        self.temperature = None
        self.setpoint = None

        self.supports_temperature = \
            group_address_temperature is not None
        self.supports_setpoint = \
            group_address_setpoint is not None

    @classmethod
    def from_config(cls, xknx, name, config):
        """Initialize object from configuration structure."""
        group_address_temperature = \
            config.get('group_address_temperature')
        group_address_setpoint = \
            config.get('group_address_setpoint')
        return cls(xknx,
                   name,
                   group_address_temperature=group_address_temperature,
                   group_address_setpoint=group_address_setpoint)

    def has_group_address(self, group_address):
        """Test if device has given group address."""
        return self.group_address_temperature == group_address or \
               self.group_address_setpoint == group_address

    @asyncio.coroutine
    def set_internal_setpoint(self, setpoint):
        """Set internal value of setpoint. Call hooks if setpoint was changed."""
        if setpoint != self.setpoint:
            self.setpoint = setpoint
            yield from self.after_update()

    @asyncio.coroutine
    def set_setpoint(self, setpoint):
        """Send setpoint to KNX bus."""
        if not self.supports_setpoint:
            return
        yield from self.send(self.group_address_setpoint, DPTArray(setpoint))
        yield from self.set_internal_setpoint(setpoint)

    @asyncio.coroutine
    def process(self, telegram):
        """Process incoming telegram."""
        if telegram.group_address == self.group_address_temperature and \
                self.supports_temperature:
            yield from self._process_temperature(telegram)
        elif telegram.group_address == self.group_address_setpoint and \
                self.supports_setpoint:
            yield from self._process_setpoint(telegram)

    @asyncio.coroutine
    def _process_temperature(self, telegram):
        """Process incoming telegram for temperature."""
        if not isinstance(telegram.payload, DPTArray) \
                or len(telegram.payload.value) != 2:
            raise CouldNotParseTelegram()
        self.temperature = DPTTemperature().from_knx(
            (telegram.payload.value[0],
             telegram.payload.value[1]))
        self.last_set = time.time()
        yield from self.after_update()

    @asyncio.coroutine
    def _process_setpoint(self, telegram):
        """Process incoming telegram for setpoint."""
        if not isinstance(telegram.payload, DPTArray) \
                or len(telegram.payload.value) != 2:
            raise CouldNotParseTelegram()
        self.setpoint = DPTTemperature().from_knx(
            (telegram.payload.value[0],
             telegram.payload.value[1]))
        yield from self.after_update()

    def state_addresses(self):
        """Return group addresses which should be requested to sync state."""
        state_addresses = []
        if self.supports_temperature:
            state_addresses.append(self.group_address_temperature)
        if self.supports_setpoint:
            state_addresses.append(self.group_address_setpoint)
        return state_addresses

    def __str__(self):
        """Return object as readable string."""
        last_set_formatted = \
                datetime.datetime.fromtimestamp(
                    self.last_set).strftime('%Y-%m-%d %H:%M:%S') \
                if self.last_set else None
        return '<Climate name="{0}" ' \
               'group_address_temperature="{1}"  ' \
               'group_address_setpoint="{2}" ' \
               'temperature="{3}" last_set="{4}" />' \
               .format(self.name,
                       self.group_address_temperature,
                       self.group_address_setpoint,
                       self.temperature,
                       last_set_formatted)

    def __eq__(self, other):
        """Equal operator."""
        return self.__dict__ == other.__dict__



"""
Base class for sending a specific type of KNX/IP Packet to a KNX/IP device and wait for the corresponding answer.

Will report if the corresponding answer was not received.
"""
import asyncio
from xknx.knxip import ErrorCode

class RequestResponse():
    """Class for ending a specific type of KNX/IP Packet to a KNX/IP and wait for the corresponding answer."""

    # pylint: disable=too-many-instance-attributes

    def __init__(self, xknx, udp_client, awaited_response_class, timeout_in_seconds=5):
        """Initialize RequstResponse class."""
        self.xknx = xknx
        self.udpclient = udp_client
        self.awaited_response_class = awaited_response_class
        self.response_received_or_timeout = asyncio.Event()
        self.success = False
        self.timeout_in_seconds = timeout_in_seconds
        self.timeout_callback = None
        self.timeout_handle = None

    def create_knxipframe(self):
        """Create KNX/IP Frame object to be sent to device."""
        raise NotImplementedError('create_knxipframe has to be implemented')

    @asyncio.coroutine
    def start(self):
        """Start. Sending and waiting for answer."""
        callb = self.udpclient.register_callback(
            self.response_rec_callback, [self.awaited_response_class.service_type])

        yield from self.send_request()
        yield from self.start_timeout()
        yield from self.response_received_or_timeout.wait()
        yield from self.stop_timeout()

        self.udpclient.unregister_callback(callb)


    @asyncio.coroutine
    def send_request(self):
        """Build knxipframe (within derived class) and send via UDP."""
        knxipframe = self.create_knxipframe()
        knxipframe.normalize()

        self.udpclient.send(knxipframe)


    def response_rec_callback(self, knxipframe, _):
        """Callback for receiving answer. Checks if answer is of correct type."""
        if not isinstance(knxipframe.body, self.awaited_response_class):
            self.xknx.logger.warning("Cant understand knxipframe")
            return

        self.response_received_or_timeout.set()

        if knxipframe.body.status_code == ErrorCode.E_NO_ERROR:
            self.success = True
            self.on_success_hook(knxipframe)
        else:
            self.on_error_hook(knxipframe)


    def on_success_hook(self, knxipframe):
        """Hook for having received a valid answer. May be overwritten in derived class."""
        pass


    def on_error_hook(self, knxipframe):
        """Hook for having received an invalid answer. May be overwritten in derived class."""
        # pylint: disable=no-self-use
        self.xknx.logger.warning("Error: reading rading group address from KNX bus failed: %s", knxipframe.body.status_code)


    def timeout(self):
        """Handle timeout."""
        self.response_received_or_timeout.set()


    @asyncio.coroutine
    def start_timeout(self):
        """Start timeout."""
        self.timeout_handle = self.xknx.loop.call_later(
            self.timeout_in_seconds, self.timeout)

    @asyncio.coroutine
    def stop_timeout(self):
        """Stop timeout."""
        self.timeout_handle.cancel()



"""
Module for Serialization and Deserialization of a KNX Disconnect Response information.

Disconnect requests are used to disconnect a tunnel from a KNX/IP device.
With a Disconnect Response the receiving party acknowledges the valid processing of the request.
"""
from xknx.exceptions import CouldNotParseKNXIP
from .knxip_enum import KNXIPServiceType
from .body import KNXIPBody
from .error_code import ErrorCode

class DisconnectResponse(KNXIPBody):
    """Representation of a KNX Disconnect Response."""
    # pylint: disable=too-many-instance-attributes

    service_type = KNXIPServiceType.DISCONNECT_RESPONSE

    def __init__(self):
        """DisconnectResponse __init__ object."""
        super(DisconnectResponse, self).__init__()

        self.communication_channel_id = 1
        self.status_code = ErrorCode.E_NO_ERROR

    def calculated_length(self):
        """Get length of KNX/IP body."""
        return 2

    def from_knx(self, raw):
        """Parse/deserialize from KNX/IP raw data."""
        def info_from_knx(info):
            """Parse info bytes."""
            if len(info) < 2:
                raise CouldNotParseKNXIP("Disconnect info has wrong length")
            self.communication_channel_id = info[0]
            self.status_code = ErrorCode(info[1])
            return 2
        pos = info_from_knx(raw)
        return pos

    def to_knx(self):
        """Serialize to KNX/IP raw data."""
        def info_to_knx():
            """Serialize information bytes."""
            info = []
            info.append(self.communication_channel_id)
            info.append(self.status_code.value)
            return info
        data = []
        data.extend(info_to_knx())
        return data

    def __str__(self):
        """Return object as readable string."""
        return '<DisconnectResponse CommunicationChannelID="{0}" ' \
            'status_code="{1}" />'.format(
                self.communication_channel_id,
                self.status_code)
